import pprint


# https://stackoverflow.com/questions/44356693/pprint-with-custom-float-formats
class FormatPrinter(pprint.PrettyPrinter):
    def __init__(self, formats):
        super().__init__()
        self.formats = formats

    def format(self, obj, ctx, maxlvl, lvl):
        if type(obj) in self.formats:
            return self.formats[type(obj)] % obj, 1, 0
        return pprint.PrettyPrinter.format(self, obj, ctx, maxlvl, lvl)


# https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook # fmt: skip
def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter



from __future__ import annotations

from copy import copy

__all__ = ("Pipeline",)

import time
from functools import partial
from typing import Any, Callable, NamedTuple

import jax.numpy as jnp
import jaxopt
import numpy.random as npr
import optax
import pyhf
import relaxed
from chex import Array
from jax import jit
from sklearn.model_selection import train_test_split

from neos.utils import isnotebook

in_jupyter = isnotebook()
if in_jupyter:
    from IPython import display


@partial(
    jit,
    static_argnames=[
        "model",
        "test_stat",
        "return_mle_pars",
        "return_constrained_pars",
    ],
)  # forward pass
def hypotest(
    test_poi: float,
    data: Array,
    model: pyhf.Model,
    lr: float,
    bonly_pars: Array,
    test_stat: str = "qmu",
    return_constrained_pars: bool = False,
) -> tuple[Array, Array] | Array:
    # hard-code 1 as inits for now
    # TODO: need to parse different inits for constrained and global fits
    init_pars = jnp.asarray(model.config.suggested_init())[
        model.config.par_slice("correlated_bkg_uncertainty")
    ]
    conditional_pars = relaxed.mle.fixed_poi_fit(
        data, model, poi_condition=test_poi, init_pars=init_pars, lr=lr
    )
    mle_pars = bonly_pars
    if test_stat == "qmu":
        profile_likelihood = -2 * (
            model.logpdf(conditional_pars, data)[0] - model.logpdf(mle_pars, data)[0]
        )

        poi_hat = mle_pars[model.config.poi_index]
        tstat = jnp.where(poi_hat < test_poi, profile_likelihood, 0.0)

    CLsb = 1 - pyhf.tensorlib.normal_cdf(jnp.sqrt(tstat))
    altval = 0.0
    CLb = 1 - pyhf.tensorlib.normal_cdf(altval)
    CLs = CLsb / CLb
    if return_constrained_pars:
        return CLs, conditional_pars
    else:
        return CLs


class Pipeline(NamedTuple):
    """Class to compose the pipeline for training a learnable summary statistic."""

    yields_from_pars: Callable[..., tuple[Array, ...]]
    model_from_yields: Callable[..., pyhf.Model]
    init_pars: Array
    nn: Callable[..., Any] | None = None
    data: Array | None = None
    yield_kwargs: dict[str, Any] | None = None
    nuisance_parname: str = "correlated_bkg_uncertainty"
    random_state: int = 0
    num_epochs: int = 20
    batch_size: int = 500
    learning_rate: float = 0.001
    optimizer: str = "adam"
    loss: Callable[[dict], float] = lambda x: x["CLs"]
    test_size: float = 0.2
    per_epoch_callback: Callable = lambda x: None
    first_epoch_callback: Callable = lambda x: None
    last_epoch_callback: Callable = lambda x: None
    post_training_callback: Callable = lambda x: None
    plot_setup: Callable = lambda x: None
    possible_metrics: tuple[str, ...] = (
        "CLs",
        "mu_uncert",
        "1-pull_width**2",
        "gaussianity",
    )
    animate: bool = True
    plot_name: str = "neos_demo.png"
    animation_name: str = "neos_demo.gif"
    plot_title: str | None = None
    plot_kwargs: dict | None = None
    float_bin_edges: bool = False
    return_pars: bool = (False,)
    data_generator: dict[str, Any] | None = (None,)

    def run(self):
        pyhf.set_backend("jax", default=True)

        def pipeline(pars, data, test=False):
            if self.float_bin_edges:
                nn_pars, bins = pars
            elif not self.float_bin_edges:
                nn_pars = pars
                bins = self.yield_kwargs["bins"]
            else:
                raise ValueError("float_bin_edges must be either True or False")

            ykw = copy(self.yield_kwargs) if self.yield_kwargs is not None else {}
            del ykw["bins"]
            ykw["use_kde"] = not test
            yields = self.yields_from_pars(nn_pars, data, self.nn, bins=bins, **ykw)
            model = self.model_from_yields(*yields)  # , validate = test)
            state: dict[str, Any] = {}
            state["yields"] = yields
            bonly_pars = (
                jnp.asarray(model.config.suggested_init())
                .at[model.config.poi_index]
                .set(0.0)
            )
            data_hf = jnp.asarray(model.expected_data(bonly_pars))
            state["CLs"], constrained = hypotest(
                1.0,
                data_hf,
                model,
                return_constrained_pars=True,
                bonly_pars=bonly_pars,
                lr=1e-2,
            )
            uncerts = relaxed.cramer_rao_uncert(model, bonly_pars, data_hf)
            state["mu_uncert"] = uncerts[model.config.poi_index]
            pull_width = uncerts[model.config.par_slice(self.nuisance_parname)][0]
            state["pull_width"] = pull_width
            state["1-pull_width**2"] = (1 - pull_width) ** 2
            # state["gaussianity"] = relaxed.gaussianity(model, bonly_pars, data, rng_key=PRNGKey(self.random_state))
            state["pull"] = jnp.array(
                [
                    (constrained - jnp.array(model.config.suggested_init()))[
                        model.config.par_order.index(k)
                    ]
                    / model.config.param_set(k).width()[0]
                    for k in model.config.par_order
                    if model.config.param_set(k).constrained
                ]
            )[0]
            state["data"] = data
            state["pars"] = pars
            state["nn"] = self.nn
            loss = self.loss(state)
            del state["data"]
            del state["nn"]
            state["loss"] = loss
            return loss, state

        if self.data is None:
            data = self.data_generator["function"](self.data_generator["args"])
        elif self.data is not None:
            data = self.data
        else:
            raise ValueError("data must be either None or an array")
        split = train_test_split(
            *data, test_size=self.test_size, random_state=self.random_state
        )
        train, test = split[::2], split[1::2]

        num_train = train[0].shape[0]
        num_complete_batches, leftover = divmod(num_train, self.batch_size)
        num_batches = num_complete_batches + bool(leftover)

        # batching mechanism
        def data_stream():
            rng = npr.RandomState(self.random_state)
            while True:
                perm = rng.permutation(num_train)
                for i in range(num_batches):
                    batch_idx = perm[i * self.batch_size : (i + 1) * self.batch_size]
                    yield [points[batch_idx] for points in train]

        batches = data_stream()

        solver = jaxopt.OptaxSolver(
            fun=pipeline, opt=optax.adam(self.learning_rate), has_aux=True, jit=True
        )
        state = solver.init_state(init_params=self.init_pars)
        params = self.init_pars

        pkw = copy(self.plot_kwargs) if self.plot_kwargs is not None else {}
        pkw["histlim"] = 100 if "histlim" not in pkw else pkw["histlim"]
        plot_kwargs = self.plot_setup(self)

        epoch_grid = jnp.linspace(0, self.num_epochs, num_batches * self.num_epochs)
        metrics = {
            "CLs": [],
            "mu_uncert": [],
            "1-pull_width**2": [],
            "loss": [],
            "test_loss": [],
            "pull": [],
            "epoch_grid": [],
            # "pars": {},
        }
        if self.return_pars:
            metrics["pars"] = {}
        metric_keys = list(metrics.keys())
        for epoch_num in range(self.num_epochs):
            print(f"epoch {epoch_num}/{self.num_epochs}: {num_batches} batches")
            for batch_num in range(num_batches):
                print(f"batch {batch_num+1}/{num_batches}:")
                batch_data = next(batches)
                start = time.perf_counter()
                params, state = solver.update(
                    params=params, state=state, data=batch_data
                )
                end = time.perf_counter()
                test_loss, test_metrics = pipeline(pars=params, data=test, test=True)
                t = end - start
                for key in test_metrics:
                    if key == "loss":
                        metrics["loss"].append(state.aux[key])
                        metrics["test_loss"].append(test_loss)
                    elif key == "pars":
                        continue
                    else:
                        if key in metric_keys:
                            metrics[key].append(test_metrics[key])
                        else:
                            metrics[key] = test_metrics[key]

                if in_jupyter:
                    display.clear_output(wait=True)
                batch_loss = state.aux["loss"]
                print(f"epoch {epoch_num}/{self.num_epochs}: {num_batches} batches")
                print(f"batch {batch_num+1}/{num_batches} took {t:.4f}s.")
                print()
                print(f"batch loss: {batch_loss:.3g}")
                print("metrics evaluated on test set:")
                for k, v in test_metrics.items():
                    if k == "yields":
                        print("yields:")
                        for label, yields in zip(["s", "b", "bup", "bdown"], v):
                            print("  ", end="")
                            print(f"{label} = [", end="")
                            for i, count in enumerate(yields):
                                if i == len(yields) - 1:
                                    print(f"{count:.3g}", end="")
                                else:
                                    print(f"{count:.3g}, ", end="")
                            print("]")

                    elif k == "pars":
                        continue
                    else:
                        print(f"{k} = {v:.3g}")
                print()

                if batch_num + epoch_num == 0:
                    if self.animate:
                        plot_kwargs["camera"] = self.first_epoch_callback(
                            params,
                            this_batch=test,
                            metrics=metrics,
                            maxN=self.num_epochs,
                            batch_num=batch_num,
                            epoch_grid=epoch_grid,
                            nn=self.nn,
                            **self.yield_kwargs,
                            **plot_kwargs,
                            **pkw,
                        )
                elif batch_num + epoch_num == num_batches - 1 + self.num_epochs - 1:
                    plot_kwargs["camera"] = self.last_epoch_callback(
                        params,
                        this_batch=test,
                        metrics=metrics,
                        maxN=self.num_epochs,
                        batch_num=batch_num + (epoch_num * num_batches),
                        epoch_grid=epoch_grid,
                        nn=self.nn,
                        pipeline=self,
                        **self.yield_kwargs,
                        **plot_kwargs,
                        **pkw,
                    )
                else:
                    if self.animate:
                        plot_kwargs["camera"] = self.per_epoch_callback(
                            params,
                            this_batch=test,
                            metrics=metrics,
                            maxN=self.num_epochs,
                            batch_num=batch_num + (epoch_num * num_batches),
                            nn=self.nn,
                            epoch_grid=epoch_grid,
                            **self.yield_kwargs,
                            **plot_kwargs,
                            **pkw,
                        )
            if "pars" in metrics:
                metrics["pars"]["post-epoch-" + str(epoch_num)] = test_metrics["pars"]
        if self.animate:
            ani = plot_kwargs["camera"].animate()
            ani.save(f"{self.animation_name}", writer="imagemagick", fps=10)
            return ani, metrics
        if self.return_pars:
            return metrics
        else:
            return test_metrics



__all__ = (
    "generate_data",
    "make_model",
    "nn_summary_stat",
    "plot",
    "plot_setup",
    "first_epoch",
    "last_epoch",
    "per_epoch",
)

import jax
import jax.numpy as jnp
import jax.scipy as jsp
import matplotlib.pyplot as plt
import numpy as np
import pyhf
import relaxed
from celluloid import Camera
from jax.random import PRNGKey, multivariate_normal


def generate_data(
    rng=0,
    num_points=10000,
    sig_mean=(-1, 1),
    bup_mean=(2.5, 2),
    bdown_mean=(-2.5, -1.5),
    b_mean=(1, -1),
):
    sig = multivariate_normal(
        PRNGKey(rng),
        jnp.asarray(sig_mean),
        jnp.asarray([[1, 0], [0, 1]]),
        shape=(num_points,),
    )
    bkg_up = multivariate_normal(
        PRNGKey(rng),
        jnp.asarray(bup_mean),
        jnp.asarray([[1, 0], [0, 1]]),
        shape=(num_points,),
    )
    bkg_down = multivariate_normal(
        PRNGKey(rng),
        jnp.asarray(bdown_mean),
        jnp.asarray([[1, 0], [0, 1]]),
        shape=(num_points,),
    )

    bkg_nom = multivariate_normal(
        PRNGKey(rng),
        jnp.asarray(b_mean),
        jnp.asarray([[1, 0], [0, 1]]),
        shape=(num_points,),
    )
    return sig, bkg_nom, bkg_up, bkg_down


def make_model(s, b_nom, b_up, b_down, validate=False):
    m = {
        "channels": [
            {
                "name": "singlechannel",
                "samples": [
                    {
                        "name": "signal",
                        "data": s,
                        "modifiers": [
                            {"name": "mu", "type": "normfactor", "data": None},
                        ],
                    },
                    {
                        "name": "background",
                        "data": b_nom,
                        "modifiers": [
                            {
                                "name": "correlated_bkg_uncertainty",
                                "type": "histosys",
                                "data": {"hi_data": b_up, "lo_data": b_down},
                            },
                        ],
                    },
                ],
            },
        ],
    }
    return pyhf.Model(m, validate=validate)


def nn_summary_stat(
    pars,
    data,
    nn,
    bandwidth,
    bins,
    reflect=False,
    sig_scale=2,
    bkg_scale=10,
    LUMI=10,
    return_preds=False,
    use_kde=True,
    use_list=False,
):
    s_data, b_nom_data, b_up_data, b_down_data = data

    preds = nn_s, nn_b_nom, nn_b_up, nn_b_down = (
        nn(pars, s_data).ravel(),
        nn(pars, b_nom_data).ravel(),
        nn(pars, b_up_data).ravel(),
        nn(pars, b_down_data).ravel(),
    )

    num_points = len(s_data)
    if use_kde:
        yields = s, b_nom, b_up, b_down = [
            relaxed.hist(nn_s, bins, bandwidth, reflect_infinities=reflect)
            * sig_scale
            / num_points
            * LUMI,
            relaxed.hist(nn_b_nom, bins, bandwidth, reflect_infinities=reflect)
            * bkg_scale
            / num_points
            * LUMI,
            relaxed.hist(nn_b_up, bins, bandwidth, reflect_infinities=reflect)
            * bkg_scale
            / num_points
            * LUMI,
            relaxed.hist(
                nn_b_down,
                bins,
                bandwidth,
                reflect_infinities=reflect,
            )
            * bkg_scale
            / num_points
            * LUMI,
        ]
    elif not use_kde:
        yields = [
            np.histogram(nn_s, bins=bins)[0] * sig_scale / num_points * LUMI,
            np.histogram(nn_b_nom, bins=bins)[0] * bkg_scale / num_points * LUMI,
            np.histogram(nn_b_up, bins=bins)[0] * bkg_scale / num_points * LUMI,
            np.histogram(nn_b_down, bins=bins)[0] * bkg_scale / num_points * LUMI,
        ]
        if use_list:
            yields = [y.tolist() for y in yields]
    else:
        raise ValueError("use_kde must be True or False")
    if return_preds:
        return yields, preds
    return yields


def make_kde(data, bw):
    @jax.jit
    def get_kde(x):
        return jnp.mean(
            jsp.stats.norm.pdf(x, loc=data.reshape(-1, 1), scale=bw), axis=0
        )

    return get_kde


def bar_plot(
    ax, data, colors=None, total_width=0.8, single_width=1, legend=True, bins=None
):
    """Draws a bar plot with multiple bars per data point.

    Parameters
    ----------
    ax : matplotlib.pyplot.axis
        The axis we want to draw our plot on.

    data: dictionary
        A dictionary containing the data we want to plot. Keys are the names of the
        data, the items is a list of the values.

        Example:
        data = {
            "x":[1,2,3],
            "y":[1,2,3],
            "z":[1,2,3],
        }

    colors : array-like, optional
        A list of colors which are used for the bars. If None, the colors
        will be the standard matplotlib color cyle. (default: None)

    total_width : float, optional, default: 0.8
        The width of a bar group. 0.8 means that 80% of the x-axis is covered
        by bars and 20% will be spaces between the bars.

    single_width: float, optional, default: 1
        The relative width of a single bar within a group. 1 means the bars
        will touch eachother within a group, values less than 1 will make
        these bars thinner.

    legend: bool, optional, default: True
        If this is set to true, a legend will be added to the axis.
    """

    # Check if colors where provided, otherwhise use the default color cycle
    if colors is None:
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]

    # Number of bars per group
    n_bars = len(data)

    # The width of a single bar
    bar_width = total_width / n_bars

    # List containing handles for the drawn bars, used for the legend
    bars = []

    # Iterate over all data
    for i, (_, values) in enumerate(data.items()):
        # The offset in x direction of that bar
        x_offset = (i - n_bars / 2) * bar_width + bar_width / 2

        # Draw a bar for every value of that type
        for x, y in enumerate(values):
            bar = ax.bar(
                x + x_offset,
                y,
                width=bar_width * single_width,
                color=colors[i % len(colors)],
            )

        # Add a handle to the last drawn bar, which we'll need for the legend
        bars.append(bar[0])

    labels = [f"[{a:.1g},{b:.1g}]" for a, b in zip(bins[:-1], bins[1:])]
    ax.set_xticks(range(len(labels)))
    ax.set_xticklabels(labels)

    return bars, data.keys()


def plot(
    network,
    axs,
    axins,
    metrics,
    maxN,
    this_batch,
    epoch_grid,
    nn,
    bins,
    bandwidth,
    batch_num,
    legend=False,
    reflect=False,
    histlim=55,
):
    if "Likelihood scan" in axs:
        ax = axs["Likelihood scan"]
        import cabinetry

        model = make_model(*metrics["yields"])
        bonly_pars = (
            jnp.asarray(model.config.suggested_init())
            .at[model.config.poi_index]
            .set(0.0)
            .tolist()
        )
        data_hf = model.expected_data(bonly_pars)
        scan_results = cabinetry.fit.scan(
            model, data_hf, "mu", par_bounds=[[-2, 10], [-2, 10]]
        )
        cabinetry.visualize.scan(scan_results, existing_ax=ax, legend=legend)
    if "Expected limits" in axs:
        # if batch_num != 0:
        ax = axs["Expected limits"]
        import cabinetry

        model = make_model(*metrics["yields"])
        bonly_pars = (
            jnp.asarray(model.config.suggested_init())
            .at[model.config.poi_index]
            .set(0.0)
            .tolist()
        )
        data_hf = model.expected_data(bonly_pars)

        limit_results = cabinetry.fit.limit(model, data_hf, maxiter=1000)
        cabinetry.visualize.limit(limit_results, existing_ax=ax, legend=legend)

    if "Data space" in axs:
        ax = axs["Data space"]
        g = np.mgrid[-5:5:101j, -5:5:101j]
        if jnp.inf in bins:
            levels = bins[1:-1]  # infinite
        else:
            levels = bins
        ax.contourf(
            g[0],
            g[1],
            nn(network, np.moveaxis(g, 0, -1)).reshape(101, 101, 1)[:, :, 0],
            levels=levels,
            cmap="binary",
        )
        ax.contour(
            g[0],
            g[1],
            nn(network, np.moveaxis(g, 0, -1)).reshape(101, 101, 1)[:, :, 0],
            colors="w",
            levels=levels,
        )
        sig, bkg_nom, bkg_up, bkg_down = this_batch
        # should definitely not have to repeat this every time lmao
        ax.scatter(sig[:, 0], sig[:, 1], alpha=0.3, c="C9", label="signal")
        ax.scatter(
            bkg_up[:, 0],
            bkg_up[:, 1],
            alpha=0.1,
            c="orangered",
            marker=6,
            label="bkg up",
        )
        ax.scatter(
            bkg_down[:, 0],
            bkg_down[:, 1],
            alpha=0.1,
            c="gold",
            marker=7,
            label="bkg down",
        )
        ax.scatter(bkg_nom[:, 0], bkg_nom[:, 1], alpha=0.3, c="C1", label="bkg")

        ax.set_xlim(-5, 5)
        ax.set_ylim(-5, 5)
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        if legend:
            ax.legend(fontsize="x-small", loc="upper right", fancybox=True)

        # ax.axis("off")
    x_grid = epoch_grid[: batch_num + 1]
    if "Losses" in axs:
        ax = axs["Losses"]
        ax.plot(
            epoch_grid[: batch_num + 1],
            metrics["loss"],
            c="C9",
            linewidth=2.0,
            label=r"train",
        )
        ax.plot(
            epoch_grid[: batch_num + 1],
            metrics["test_loss"],
            c="C4",
            linewidth=2.0,
            label=r"test",
        )
        ax.set_yscale("log")
        ax.set_xlim(0, maxN)
        ax.set_xlabel("epoch")
        ax.set_ylabel(r"loss value")
        if legend:
            ax.legend(fontsize="x-small", loc="upper right", fancybox=True)

    if "Metrics" in axs:
        ax = axs["Metrics"]
        ax.plot(
            x_grid,
            metrics["1-pull_width**2"],
            c="slategray",
            linewidth=2.0,
            label=r"$(1-\sigma_{\mathsf{nuisance}})^2$",
            # linestyle=":"
        )
        ax.plot(
            x_grid,
            np.array(metrics["pull"]) ** 2,
            c="C2",
            linewidth=2.0,
            label=r"(nuisance pull)$^2$",
            # linestyle=':'
        )
        ax.plot(
            x_grid,
            metrics["mu_uncert"],
            c="steelblue",
            linewidth=2.0,
            label=r"$\sigma_\mu$",
        )
        ax.plot(x_grid, metrics["CLs"], c="C9", linewidth=2, label=r"$CL_s$")
        ax.set_ylim(1e-7, 1e-0)
        ax.set_xlim(0, maxN)
        ax.set_xlabel("epoch")
        ax.set_yscale("log")
        ax.set_ylabel(r"metric value (on test set)")
        if legend:
            ax.legend(fontsize="x-small", loc="upper right", fancybox=True)
    if "Histogram model" in axs:
        ax = axs["Histogram model"]
        s, b, bup, bdown = metrics["yields"]

        if jnp.inf in bins:
            noinf = bins[1:-1]
            bin_width = 1 / (len(noinf) - 1)
            centers = noinf[:-1] + np.diff(noinf) / 2.0
            centers = jnp.array([noinf[0] - bin_width, *centers, noinf[-1] + bin_width])

        dct = {
            "signal": s,
            "bkg up": bup,
            "bkg": b,
            "bkg down": bdown,
        }

        a, b = bar_plot(
            ax,
            dct,
            colors=["C9", "orangered", "C1", "gold"],
            total_width=0.8,
            single_width=1,
            legend=legend,
            bins=bins,
        )
        ax.set_ylabel("frequency")
        ax.set_xlabel("interval over nn output")
        ax.set_ylim(0, histlim)
        # ax.axis("off")

        if legend:
            # Draw legend if we need
            # ax.legend(a, b, fontsize="x-small")
            if jnp.inf in bins:
                width = jnp.diff(noinf)[0]
            else:
                width = jnp.diff(bins)[0]
            xlim = (
                [(width / 2) - (1.1 * bandwidth), (width / 2) + (1.1 * bandwidth)]
                if (width / 2) - bandwidth < 0
                else [-width / 3, width + width / 3]
            )
            axins.stairs([1], [0, width], color="C1", alpha=0.6)
            y = jnp.linspace(xlim[0], xlim[1], 300)
            demo = jsp.stats.norm.pdf(y, loc=width / 2, scale=bandwidth)
            axins.plot(
                y, demo / max(demo), color="C0", linestyle="dashed", label="kernel"
            )
            # draw two vertical lines at ((width/2)-bandwidth)/2 and ((width/2)+bandwidth)/2
            axins.vlines(
                [(width / 2) - bandwidth, (width / 2) + bandwidth],
                0,
                1,
                colors="black",
                linestyles="dotted",
                label=r"$\pm$bandwidth",
                alpha=0.9,
            )
            # write text in the middle of the vertical lines with the value of the bandwidth
            ratio = bandwidth / width
            axins.text(
                width / 2,
                -0.3,
                r"$\mathsf{\frac{bandwidth}{bin\,width}}=$" + f"{ratio:.2f}",
                ha="center",
                va="center",
                size="x-small",
                alpha=0.9,
            )

            axins.set_xlim(*xlim)

            handles, labels = a, list(b)  # ax.get_legend_handles_labels()
            handles1, labels1 = axins.get_legend_handles_labels()
            ax.legend(
                handles + handles1,
                labels + labels1,
                loc="upper right",
                fontsize="x-small",
                fancybox=True,
            )

    if "Nuisance pull" in axs:
        ax = axs["Nuisance pull"]

        pulls = metrics["pull"]
        pullerr = metrics["pull_width"]

        ax.set_ylabel(r"$(\theta - \hat{\theta})\,/ \Delta \theta$", fontsize=18)

        # draw the +/- 2.0 horizontal lines
        ax.hlines([-2, 2], -0.5, len(pulls) - 0.5, colors="black", linestyles="dotted")
        # draw the +/- 1.0 horizontal lines
        ax.hlines([-1, 1], -0.5, len(pulls) - 0.5, colors="black", linestyles="dashdot")
        # draw the +/- 2.0 sigma band
        ax.fill_between([-0.5, len(pulls) - 0.5], [-2, -2], [2, 2], facecolor="yellow")
        # drawe the +/- 1.0 sigma band
        ax.fill_between([-0.5, len(pulls) - 0.5], [-1, -1], [1, 1], facecolor="green")
        # draw a horizontal line at pull=0.0
        ax.hlines([0], -0.5, len(pulls) - 0.5, colors="black", linestyles="dashed")

        ax.scatter(range(len(pulls)), pulls, color="black")
        # and their uncertainties
        ax.errorbar(
            range(len(pulls)),
            pulls,
            color="black",
            xerr=0,
            yerr=pullerr,
            marker=".",
            fmt="none",
        )
    if "Example KDE" in axs:
        ax = axs["Example KDE"]
        b_data = bkg_nom
        d = np.array(nn(network, b_data).ravel().tolist())
        kde = make_kde(d, bandwidth)
        yields = b
        ls = [-1, 2]
        x = np.linspace(ls[0], ls[1], 300)
        db = jnp.array(jnp.diff(bins), float)  # bin spacing
        yields = yields / db / yields.sum(axis=0)  # normalize to bin width
        if jnp.inf in bins:
            pbins = [ls[0], *noinf, ls[1]]
        else:
            pbins = bins
        ax.stairs(yields, pbins, label="KDE hist", color="C1")
        if reflect:
            ax.plot(x, 2 * jnp.abs(kde(x)), label="KDE", color="C0")
        else:
            ax.plot(x, kde(x), label="KDE", color="C0")

        ax.set_xlim(*ls)

        # rug plot of the data
        ax.plot(
            d,
            jnp.zeros_like(d) - 0.01,
            "|",
            linewidth=3,
            alpha=0.4,
            color="black",
            label="data",
        )

        if legend:
            if jnp.inf in bins:

                width = jnp.diff(noinf)[0]
            else:
                width = jnp.diff(bins)[0]
            xlim = (
                [(width / 2) - (1.1 * bandwidth), (width / 2) + (1.1 * bandwidth)]
                if (width / 2) - bandwidth < 0
                else [-width / 3, width + width / 3]
            )
            axins.stairs([1], [0, width], color="C1")
            y = jnp.linspace(xlim[0], xlim[1], 300)
            demo = jsp.stats.norm.pdf(y, loc=width / 2, scale=bandwidth)
            axins.plot(
                y, demo / max(demo), color="C0", linestyle="dashed", label="kernel"
            )
            # draw two vertical lines at ((width/2)-bandwidth)/2 and ((width/2)+bandwidth)/2
            axins.vlines(
                [(width / 2) - bandwidth, (width / 2) + bandwidth],
                0,
                1,
                colors="black",
                linestyles="dotted",
                label=r"$\pm$bandwidth",
            )
            # write text in the middle of the vertical lines with the value of the bandwidth
            ratio = bandwidth / width
            axins.text(
                width / 2,
                -0.3,
                r"$\mathsf{\frac{bandwidth}{bin\,width}}=$" + f"{ratio:.2f}",
                ha="center",
                va="center",
                size="x-small",
            )

            axins.set_xlim(*xlim)

            handles, labels = ax.get_legend_handles_labels()
            handles1, labels1 = axins.get_legend_handles_labels()
            ax.legend(
                handles + handles1,
                labels + labels1,
                loc="upper right",
                fontsize="x-small",
                fancybox=True,
            )


def first_epoch(
    network,
    camera,
    axs,
    axins,
    metrics,
    maxN,
    this_batch,
    epoch_grid,
    batch_num,
    nn,
    bins,
    bandwidth,
    histlim,
    **kwargs,
):
    plot(
        axs=axs,
        axins=axins,
        network=network,
        metrics=metrics,
        maxN=maxN,
        this_batch=this_batch,
        batch_num=batch_num,
        epoch_grid=epoch_grid,
        nn=nn,
        bins=bins,
        bandwidth=bandwidth,
        legend=True,
        histlim=histlim,
    )
    plt.tight_layout()
    if camera is not None:
        camera.snap()
    return camera


def last_epoch(
    network,
    camera,
    axs,
    axins,
    metrics,
    maxN,
    this_batch,
    batch_num,
    epoch_grid,
    nn,
    bins,
    bandwidth,
    pipeline,
    histlim,
    **kwargs,
):
    plot(
        axs=axs,
        axins=axins,
        network=network,
        metrics=metrics,
        maxN=maxN,
        this_batch=this_batch,
        epoch_grid=epoch_grid,
        batch_num=batch_num,
        nn=nn,
        bins=bins,
        bandwidth=bandwidth,
        histlim=histlim,
        legend=False,
    )
    plt.tight_layout()
    if camera is not None:
        camera.snap()

    defaults = {
        "layout": [
            ["Data space", "Histogram model"],
            ["Losses", "Metrics"],
        ]
    }
    if pipeline.plot_kwargs is not None:
        for k in defaults:
            if k in pipeline.plot_kwargs:
                defaults[k] = pipeline.plot_kwargs[k]
    fig2, axs2 = plt.subplot_mosaic(defaults["layout"])

    for label, ax in axs2.items():
        ax.set_title(label, fontstyle="italic")
    if "Histogram model" in axs2:
        axins2 = axs2["Histogram model"].inset_axes([0.33, 0.79, 0.3, 0.2])
        axins2.axis("off")
    else:
        axins2 = None
    plot(
        axs=axs2,
        axins=axins2,
        network=network,
        metrics=metrics,
        maxN=maxN,
        this_batch=this_batch,
        nn=nn,
        epoch_grid=epoch_grid,
        batch_num=batch_num,
        bins=bins,
        bandwidth=bandwidth,
        legend=True,
        histlim=histlim,
    )
    plt.tight_layout()
    fig2.savefig(f"{pipeline.plot_name}")
    return camera


def per_epoch(
    network,
    camera,
    axs,
    axins,
    metrics,
    maxN,
    this_batch,
    epoch_grid,
    batch_num,
    nn,
    bins,
    bandwidth,
    histlim,
    **kwargs,
):
    plot(
        axs=axs,
        axins=axins,
        network=network,
        metrics=metrics,
        maxN=maxN,
        this_batch=this_batch,
        epoch_grid=epoch_grid,
        batch_num=batch_num,
        nn=nn,
        bins=bins,
        bandwidth=bandwidth,
        legend=False,
        histlim=histlim,
    )
    plt.tight_layout()
    if camera is not None:
        camera.snap()
    return camera


def plot_setup(pipeline):
    plt.style.use("default")

    defaults = {
        "layout": [
            ["Data space", "Histogram model"],
            ["Losses", "Metrics"],
        ]
    }
    if pipeline.plot_kwargs is not None:
        for k in defaults:
            if k in pipeline.plot_kwargs:
                defaults[k] = pipeline.plot_kwargs[k]

    plt.rcParams.update(
        {
            "axes.labelsize": 13,
            "axes.linewidth": 1.2,
            "xtick.labelsize": 13,
            "ytick.labelsize": 13,
            "figure.figsize": [10.0, 10.0],
            "font.size": 13,
            "xtick.major.size": 3,
            "ytick.major.size": 3,
            "legend.fontsize": 11,
            "legend.fancybox": True,
        }
    )

    plt.rc("figure", dpi=150)

    fig, axs = plt.subplot_mosaic(defaults["layout"])

    for label, ax in axs.items():
        ax.set_title(label, fontstyle="italic")
    # axs["Example KDE"].set_title("Example KDE (nominal bkg)", fontstyle="italic")
    if "Histogram model" in axs:
        axins = axs["Histogram model"].inset_axes([0.33, 0.79, 0.3, 0.2])
        axins.axis("off")
        axins_cpy = axins
    else:
        axins = axins_cpy = None
    ax_cpy = axs
    if pipeline.animate:
        camera = Camera(fig)
    else:
        camera = None
    plt.suptitle(pipeline.plot_title, fontsize="x-large")
    return dict(
        camera=camera, axs=axs, axins=axins, ax_cpy=ax_cpy, axins_cpy=axins_cpy, fig=fig
    )
